local gamepad = require "gamepad"

-- Libraries
local rs = require "libs/resolution_solution"
local anim8 = require "libs/anim8"

Player = {}
Player.__index = Player

function Player.new(x,y)
  local self = setmetatable({}, Player)
  
  self.debug_on = 1
  
  -- Variables
  self.x = x
  self.y = y
  self.width = 28
  self.height = 16 -- +16 (foot shape radius)
  self.normal_x = 0 -- Collision direction x
  self.normal_y = 0  -- Collision direction y
  self.max_speed_x = 150
  self.max_speed_y = 150
  
  -- Player states
  self.in_ground = false
  self.in_ground_wait = false -- Used to check after some time if the player has left the ground
  self.in_ground_waitTime = 0
  
  self.in_air = true
  self.ascending = false
  self.descending = true
  
  self.is_moving = false
  self.idle = true
  
  -- Player's physics
  self.physics = {}
  
  self.physics.body = love.physics.newBody(World, self.x, self.y, "dynamic")
  
  -- Player's body
  self.physics.rect_shape = love.physics.newRectangleShape(0,0,self.width,self.height)
  self.physics.rect_fixture = love.physics.newFixture(self.physics.body, self.physics.rect_shape, 1)
  self.physics.rect_fixture:setUserData("player")
  
  -- Player's foot
  self.physics.foot_shape = love.physics.newCircleShape(0,self.height/2,(self.width/2+1))
  self.physics.foot_fix = love.physics.newFixture(self.physics.body, self.physics.foot_shape, 1)
  self.physics.foot_fix:setUserData("player_foot")
  -- I used a circle for the foot collision to achieve smoother player movement
  -- Note: The circle's radius is equal to the player's body width + 1 to prevent the player from getting stuck in walls. Dealing with physics challenges :/
  
  self.physics.body:setFixedRotation(true)
  self.physics.body:setMass(0,0,1,0)
  
  -- Player's graphics
  self.aim = "right" -- Direction the player is facing
  self.animation_aim = "right" -- Direction the animation sprites are facing
  
  -- This sprite sheet includes sprites facing both right and left, but I managed that through code.
  self.spriteSheet = assets.playerSheet 
  local player_grid = anim8.newGrid(16,16,
    self.spriteSheet:getWidth(),
    self.spriteSheet:getHeight())
  
  self.animations = {}
  self.animations.idle = anim8.newAnimation(player_grid("1-4",1),{0.2,0.25,0.2,0.1})
  self.animations.walk = anim8.newAnimation(player_grid("1-4",3),0.2)
  self.animations.jump = anim8.newAnimation(player_grid(3,5),0.2)
  self.animations.descending = anim8.newAnimation(player_grid(1,5),0.2)
  
  self.current_animation = self.animations.idle
  self.current_animation_name = "idle"
  return self
end 

function Player:update(dt)
  self:gamepad_control(gamepad)
  self:keyboard_control()
  
  self:velocity_limit(self.max_speed_x, self.max_speed_y)
  
  -- Update the wait time for checking if the player is on the ground.
  if self.in_ground_wait then
    self.in_ground_waitTime = self.in_ground_waitTime + dt
  end
  
  -- After 0.05 seconds off the ground, the code assumes that the player is no longer on the ground (necessary to avoid issues with collision callbacks).
  if self.in_ground_waitTime > 0.05 then
    self.in_ground_waitTime = 0 -- Reset
    self.in_ground_wait = false
    self.in_ground = false
    self.in_air = true
    -- The code assumes that the player is jumping (subject to later verification).
    self.current_animation = self.animations.jump
    self.current_animation_name = "jump"
  end
  
  -- Detecting if the player is moving.
  local dx, dy = self.physics.body:getLinearVelocity()
  if (dx > 1 or dx < -1) and self.in_ground then 
    self.is_moving = true
    self.idle = false
    self.current_animation = self.animations.walk
    self.current_animation_name = "walk"
  elseif self.is_moving then
    self.is_moving = false
    self.idle = true
  end
  
  -- Checking if the player is ascending or descending.
  local dx, dy = self.physics.body:getLinearVelocity()
  if self.in_air and dy < 0 then
    self.ascending = true
    self.descending = false
    self.current_animation = self.animations.jump
    self.current_animation_name = "jump"
  elseif self.in_air and dy > 100 then
    self.descending = true
    self.ascending = false
    self.current_animation = self.animations.descending
    self.current_animation_name = "descending"
  end
  
  
  -- Control of animation speed
  local dx, dy = self.physics.body:getLinearVelocity()
  local animation_speed = 1 -- Default
  if self.is_moving and self.in_ground then
    animation_speed = math.abs(dx)/50
  end
  
  -- Checks the animation direction
  if self.animation_aim ~= self.aim then
    for _, animation in pairs(self.animations) do 
      animation:flipH()
    end
    
    self.animation_aim = self.aim
  end
  
  self.current_animation:update(dt * animation_speed)
end 

function Player:draw()
  love.graphics.setColor(1,1,1,1)
  local x, y = self.physics.body:getPosition()
  local sprite_width, sprite_height = 16, 16
  
  self.current_animation:draw(self.spriteSheet, x,y, 0, 3,3,sprite_width/2,(sprite_height/2) - 1.5)
  
  if self.debug_on > 0 then
    self:debug(5,5,nil,1,1,true,true,true)
  end
end 

function Player:gamepad_control(gamepad)
  local speedX = 400
  -- Movement
  if type(gamepad) == "userdata" then
    if gamepad:isGamepadDown("dpright") then
      self.physics.body:applyForce(speedX,0)
      self.aim = "right"
    end
    if gamepad:isGamepadDown("dpleft") then
      self.physics.body:applyForce(-speedX,0)
      self.aim = "left"
    end
  end
end

function Player:keyboard_control()
  local speedX = 400
  -- Movement
  if love.keyboard.isDown("d") then
    self.physics.body:applyForce(speedX,0)
    self.aim = "right"
  end
  if love.keyboard.isDown("a") then
    self.physics.body:applyForce(-speedX,0)
    self.aim = "left"
  end
end

-- Gamepad Callbacks
function Player:joystickpressed(joy, button)
  if button == 1 and self.in_ground then
    local impulse_y = -320
    self.physics.body:applyLinearImpulse(0,impulse_y)
  end
end

-- Keyboard Callbacks
function Player:keypressed(key)
  if key == "space" then
    local impulse_y = -320
    self.physics.body:applyLinearImpulse(0,impulse_y)
  end
end

-- Physics Callbacks
function Player:beginContact(a,b,coll)
  local a_name = a:getUserData()
  local b_name = b:getUserData()
  self.normal_x, self.normal_y = coll:getNormal()
  
  -- Player's foot collides with the top of the ground.
  if self.normal_x == 0 and self.normal_y >= -1 and self.normal_y < 0 and a_name == "ground" and b_name == "player_foot" then
    self.in_ground = true
    self.ascending = false
    self.descending = false
    self.in_air = false
    self.current_animation = self.animations.idle
    self.current_animation_name = "idle"
    self.in_ground_wait = false
    self.in_ground_waitTime = 0
  end
end

function Player:endContact(a, b, coll)
  local a_name = a:getUserData()
  local b_name = b:getUserData()
  

  -- Player's foot stops colliding with the ground.
  if a_name == "ground" and b_name == "player_foot" then
    self.in_ground_wait = true
  end
end

function Player:preSolve(a,b,coll)
  local a_name = a:getUserData()
  local b_name = b:getUserData()
  self.normal_x, self.normal_y = coll:getNormal()
  
  -- Player's foot is colliding with the ground.
  if self.normal_x == 0 and self.normal_y >= -1 and self.normal_y < 0 and a_name == "ground" and b_name == "player_foot" then
    self.in_ground = true
    self.ascending = false
    self.descending = false
    self.in_air = false
    self.in_ground_wait = false
    self.in_ground_waitTime = 0
    
    if not self.is_moving then 
      self.current_animation = self.animations.idle
      self.current_animation_name = "idle"
    end
  end
end

function Player:postSolve(a,b,coll)
end

function Player:velocity_limit(max_speed_x)
  -- Note: Limits only the horizontal speed; the vertical is left to gravity.
  if self.in_air then 
    max_speed_x = max_speed_x * 1.5
  end
 
  local speed_x, speed_y = self.physics.body:getLinearVelocity()
 
  if math.abs(speed_x) > max_speed_x then
    speed_x = (speed_x/math.abs(speed_x)) * max_speed_x
  end

  self.physics.body:setLinearVelocity(speed_x, speed_y)
end 

function Player:debug(x,y,r,sx,sy,physics,states,info)
  local print_table = {}
  
  -- Shapes of collisions.
  if physics then
    love.graphics.setColor(0,1,1,0.5)
    love.graphics.polygon("line",
      self.physics.body:getWorldPoints(
      self.physics.rect_shape:getPoints()))
    
    love.graphics.circle("line",
      self.physics.body:getX(),
      self.physics.body:getY() + self.height/2,
      self.physics.foot_shape:getRadius())
  end

  if states then
    local states = {
      "STATES:",
      "in_ground:"..tostring(self.in_ground),
      "in_air:"..tostring(self.in_air),
      "in_ground_wait:"..tostring(self.in_ground_wait),
      "is_moving:"..tostring(self.is_moving),
      "idle:"..tostring(self.idle),
      "ascending:"..tostring(self.ascending),
      "descending:"..tostring(self.descending),
      "\n"
    }
    
    for _, str in ipairs(states) do 
      table.insert(print_table, str)
    end
  end
  if info then
    local spdX, spdY = self.physics.body:getLinearVelocity()
    
    local info = {
      "VARIABLES",
      "x:"..math.floor(self.physics.body:getX()),
      "y:"..math.floor(self.physics.body:getY()),
      "speed_x:"..math.floor(spdX),
      "speed_y:"..math.floor(spdY),
      "in_ground_waitTime:"..self.in_ground_waitTime,
      "normal_x:"..self.normal_x,
      "normal_y:"..self.normal_y,
      "aim:"..self.aim,
      "animation_aim:"..self.animation_aim,
      "current_animation:"..self.current_animation_name,
    }
    for _, str in ipairs(info) do 
      table.insert(print_table, str)
    end
  end
  
  love.graphics.setColor(1,1,1,1)
  love.graphics.print(table.concat(print_table, "\n"),x,y,r,sx,sy)
end
